"""
Devices are elements that create link between general specifications \
and electrical behavior.
For example, balun create en link between input and ouput impedances and
inductor value of the balun.
"""

from .balun import Balun
from .coupler import Coupler



from typing import Literal, List, Tuple
from pydantic import BaseModel
import numpy as np


class NDArray(np.ndarray):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, NDArray):
            return v
        if isinstance(v, np.ndarray):
            return v
        if isinstance(v, List) or isinstance(v, Tuple):
            return np.asarray(v)
        raise ValueError(f"cannot coerse input of type {type(v)} to numpy array.")


class PhysicalDimension(BaseModel):
    value: NDArray
    unit: str = ""
    scale: Literal["lin", "dB"] = "lin"

    class Config:
        arbitrary_types_allowed = True

    def dB(self):
        """
        Return the decibel value of the given imaginary number.
        """
        v = self.value
        v_db = v if self.scale == "dB" else 10 * np.log10(np.abs(v))
        out = self.__class__(value=v_db, scale="dB", unit=self.unit)
        return out

    def lin(self):
        """
        return the linear magnitude of the given magnitude in decibel
        """
        v = self.value
        v_lin = 10 ** (v / 10) if self.scale == "dB" else v
        return self.__class__(value=v_lin, scale="lin", unit=self.unit)

    def __getitem__(self, item):
        return self.__class__(
            value=[
                self.value[item],
            ],
            scale=self.scale,
            unit=self.unit,
        )

    def shape(self):
        return self.value.shape

    def __sub__(self, other):
        return operator(self, other, "-")

    def __add__(self, other):
        return operator(self, other, "+")

    def __truediv__(self, other):
        return operator(self, other, "/")

    def __mul__(self, other):
        return operator(self, other, "*")

    def __eq__(self, other):
        return (
            self.unit == other.unit
            and self.scale == other.scale
            and np.all(self.value == other.value)
        )

    def __round__(self, x):
        return self.__class__(
            value=np.round(self.value, x), scale=self.scale, unit=self.unit
        )


def operator(l_a, l_b, op):
    b = l_b.value if isinstance(l_b, PhysicalDimension) else l_b
    if op == "+":
        res = l_a.value + b
    if op == "-":
        res = l_a.value - b
    if op == "/":
        res = l_a.value / b
    if op == "*":
        res = l_a.value * b
    return l_a.__class__(value=res, scale=l_a.scale, unit=l_a.unit)



# CSM Model
"""
[WIP]
"""
import numpy as np
import matplotlib.pyplot as plt

# constants
q = 1.602176634e-19  # C
k = 1.380649e-23  # J/K
eps0 = 8.8541878128e-12  # F/m

Temp = 300  # K
N = 1e17  # atoms/cm3
t_ox = 5e-9  # m
mu_Cox = 3.45e-4  # A/V²
U_t = k * Temp / q  # V


def gamma(c_ox: float, N_atom: float):
    """
    body effect parameter
    """
    eps_si = 11.5 * eps0
    return 50 * np.sqrt(2 * q * eps_si * N_atom) / c_ox


def f_psi(psi: float, V_g: float):
    """
    function that link bulk potential and drain current
    """
    ga = gamma(mu_Cox, N)
    return (
        -0.5 * psi ** 2
        - (2 / 3) * ga * (psi ** 1.5)
        + (V_g + U_t) * psi
        + ga * U_t * (psi ** 0.5)
    )


def I_d(width: float, length: float, psi_ss, psi_sd, v_g):
    """
    drain current
    """
    beta = mu_Cox * width / length
    return beta * (f_psi(psi_sd, v_g) - f_psi(psi_ss, v_g))


def phi_b(N: float, n_i: float):
    """
    bulk potential
    """
    return U_t * np.log(N / n_i)


def V_non_eq(psi_s: float, V_g: float):
    """
    non-equilibrium voltage
    """
    ga = gamma(mu_Cox, N)
    n_i = 1e15
    phib = phi_b(N, n_i)
    return -U_t * np.log((((V_g - psi_s) / ga) ** 2 - psi_s) / U_t) + psi_s - 2 * phib


VG = np.arange(1.5, 3.1, 0.5)
psi_s = np.logspace(-11, 0, 101)

for vg in VG:
    ga = gamma(mu_Cox, N)
    psi_max = (np.sqrt(0.25 * ga ** 2 + vg) - ga / 2) ** 2
    print(f"$\gamma$: {ga}, $\psi_Smax$: {psi_max}")
    psi_swp = psi_max - psi_s
    plt.plot(V_non_eq(psi_swp, vg), psi_swp, label=f"VG={vg} V")
plt.legend()
plt.xlabel("V (B)")
plt.ylabel(r"$\psi_S$ (B)")
plt.grid()


plt.figure()
VD = np.arange(0, 2, 0.1)
VG = np.arange(0.5, 2.1, 0.25)
width = 65e-9
length = 10e-6
for vg in VG:
    plt.plot(VD, I_d(width, length, VD, 0, vg), label=f"VG={vg} V")
plt.legend()
plt.grid()



# -*- coding: utf-8 -*-
"""
Define constants and function dedicated to RF-conception
"""
import numpy as np
from passive_auto_design.units.physical_dimension import PhysicalDimension
from passive_auto_design.units.time import Frequency

# Other functions
def gamma(_z_load: float, _z0: float = 50):
    return (_z0 - _z_load) / (_z0 + _z_load)


def std_dev(measured, targeted):
    """
    return the standard deviation between an array_like of results and their references.
    """
    tmp = np.array(np.abs(gamma(measured, targeted)) ** 2)
    return np.sqrt(np.sum(tmp))


def ihsr(_s31, _s21):
    """
    Return the IHSR (Ideal Hybrid Splitting Ratio) for the given gains
    """
    if _s21 == 1j * _s31 or _s21 == -1j * _s31:
        return np.inf
    ihsr_log = 10 * np.log10(np.abs((_s21 - 1j * _s31) / (_s21 + 1j * _s31)))
    return -np.minimum(ihsr_log, -ihsr_log)


def quality_f(_z):
    """
    return the quality factor of a components
    """
    return np.imag(_z) / np.real(_z)


def reflexion_coef(_z_steps, _phi_step):
    """
    return the coefficient reflexion of a given sequence of transmission lines
    with the given_z_steps profile and equal length of _phi_steps degrees
    """
    n_step = len(_z_steps)

    z_tot = _z_steps[-1:]
    for i in range(n_step):
        z_0 = _z_steps[n_step - i - 1]
        z_tot = (
            z_0
            * (z_tot + 1j * z_0 * np.tan(_phi_step))
            / (z_0 + 1j * z_tot * np.tan(_phi_step))
        )
    return gamma(_z_steps[0], z_tot)


def transmission_coef(_z_steps, _phi_step):
    return np.sqrt(1 - reflexion_coef(_z_steps, _phi_step) ** 2)


def frac_bandwidth(f_min, f_max):
    return 100 * np.abs(f_max - f_min) / np.sqrt(f_max * f_min)


def friis(f: Frequency, gain: PhysicalDimension):
    """

    Parameters
    ----------
    f : np.Array
        List of the noise figure (in dB) of each block.
    gain : np.Array
        List of the gain of each block (in dB).


    Returns
    -------
    float
        Total noise figure of the system

    """

    m = gain.shape()[0]
    n = f.shape()[0]
    if m != n - 1:
        raise ValueError("gain should have 1 item less than noise factor f")
    g_tot = PhysicalDimension(value=[1.,], scale="lin", unit="")
    f_lin = f.lin()
    res = f_lin[0]
    for i in range(m):
        g_tot = g_tot*gain[i].lin()
        res += (f_lin[i + 1] - 1) / g_tot
    return res.dB()



# -*- coding: utf-8 -*-
"""

"""
import numpy as np
import passive_auto_design.components.lumped_element as lmp


class Transformer(lmp.LumpedElement):
    """
    Create a transformer object with the specified Inductors primary & secondary
    and calculate the associated electrical model
    if sym=True, Create a symmetrical transformer (secondary is ignored and assume egal to primary)
    """

    sym: bool

    def __init__(
        self,
        primary: lmp.Inductor,
        secondary=None,
        rho=0.0,
        eps_r=0.0,
        h_mut=0.0,
        h_gnd=0.0,
        sym=False,
    ):
        const = {
            "eps_r": eps_r,
            "rho": rho,
        }
        dim = {
            "h_mut": h_mut,
            "h_gnd": h_gnd,
        }
        for key in primary.dim.keys():
            dim["lp." + key] = primary.dim[key]
            dim["ls." + key] = primary.dim[key] if sym else secondary.dim[key]
        lmp.LumpedElement.__init__(self, dim=dim, const=const, sym=sym)

    def get_model(self):
        primary = lmp.Inductor(
            self.dim["lp.d_i"],
            self.dim["lp.n_turn"],
            self.dim["lp.width"],
            self.dim["lp.gap"],
        )
        self.model["lp"] = primary.model["ind"]
        if self.sym:
            self.model["ls"] = primary.model["ind"]
        else:
            secondary = lmp.Inductor(
                self.dim["ls.d_i"],
                self.dim["ls.n_turn"],
                self.dim["ls.width"],
                self.dim["ls.gap"],
            )
            self.model["ls"] = secondary.model["ind"]
        self.model["rp"] = self.r_geo(True)
        self.model["rs"] = self.r_geo(False)
        self.model["cm"] = self.cc_geo(True)
        self.model["cg"] = self.cc_geo(False)
        self.model["k"] = self.k_geo()
        return self.model

    def cc_geo(self, _mutual=True):
        """
        return the value of the distributed capacitance of the described transformer
        if _mutual, return the capacitance between primary and secondary
        else, return the capacitance to the ground plane
        """
        dim = self.dim
        if _mutual:
            dist = float(self.dim["h_mut"])
            d_i = np.max([dim["lp.d_i"], dim["ls.d_i"]])
            d_o = np.min(
                [
                    d_i + dim["lp.n_turn"] * dim["lp.width"],
                    d_i + dim["ls.n_turn"] * dim["ls.width"],
                ]
            )
        else:
            dist = float(self.dim["h_gnd"])
            d_i = np.min([dim["ls.d_i"], dim["lp.d_i"]])
            d_o = np.max(
                [
                    d_i + dim["lp.n_turn"] * dim["lp.width"],
                    d_i + dim["ls.n_turn"] * dim["ls.width"],
                ]
            )
        eps_r = float(self.const["eps_r"])
        area = 4 * (d_o ** 2 - d_i ** 2) * (1 + 2 * np.sqrt(2))
        cap = lmp.Capacitor(area, dist, eps_r)
        return np.abs(cap.model["cap"])

    def r_geo(self, _of_primary=True):
        """
        return the value of the resistance of the described transformer
        """
        geo = "lp." if _of_primary else "lp."
        rho = self.const["rho"]
        n_t = self.dim[geo + "n_turn"]
        l_tot = (
            8
            * np.tan(np.pi / 8)
            * n_t
            * (
                self.dim[geo + "d_i"]
                + self.dim[geo + "width"]
                + (n_t - 1) * (self.dim[geo + "width"] + self.dim[geo + "gap"])
            )
        )
        r_dc = lmp.Resistor(self.dim[geo + "width"], l_tot, rho)
        return r_dc.model["res"]

    def k_geo(self):
        """
        return the value of the coupling between the two inductors.

        """
        dim = self.dim
        c1 = dim["lp.width"] * dim["lp.n_turn"] + dim["lp.gap"] * (dim["lp.n_turn"] - 1)
        d1 = dim["lp.d_i"]
        c2 = dim["ls.width"] * dim["ls.n_turn"] + dim["ls.gap"] * (dim["ls.n_turn"] - 1)
        d2 = dim["ls.d_i"]
        return (
            0.99
            * (np.max([d1, d2]) - np.min([c1 + d1, c2 + d2]))
            / (np.max([c1 + d1, c2 + d2]) - np.min([d1, d2]))
        )



"""
Components are elements that define a link between geometrical dimensions\
and electrical behavior using physical parameter.
For example, Inductor create a link between the inner and outer diameter and
its self-inductance.
"""



# -*- coding: utf-8 -*-
"""
    Class to calculate the inductors values of an impedance transformer.
"""
import numpy as np
from scipy.optimize import minimize
from ..special import quality_f
from ..units.unit import Frequency, Impedance


class Balun:
    """
    Create a balun object
    """

    def __init__(self, _fc=1e9, _z_source=50, _z_load=50, _k=0.9):
        self.f_c = _fc
        self.z_src = _z_source
        self.z_ld = _z_load
        self.k = _k

    def __str__(self):
        message = f"target : fc={Frequency(self.f_c)}\tzs={Impedance(self.z_src)}\tzl={Impedance(self.z_ld)}"
        return message

    def design(self, XL_add=(0.0, 0.0), XS_add=(0.0, 0.0), r_serie=(0.0, 0.0)):
        """
        design an impedance transformer
        with the targeted specifications (f_targ, zl_targ, zs_targ, k)
        return the two ideal transformers solution
        """
        k = self.k
        alpha = (1 - k ** 2) / k ** 2
        q_s = -quality_f(self.z_src + 1j * np.array(XS_add))
        q_l = -quality_f(self.z_ld + 1j * np.array(XL_add))
        # assuming perfect inductor for first calculation
        r_l1, r_l2 = r_serie
        q_s_prime = q_s * np.real(self.z_src) / (np.real(self.z_src) + r_l1)
        q_l_prime = q_l * np.real(self.z_ld) / (np.real(self.z_ld) + r_l2)
        b_coeff = 2 * alpha * q_s_prime + q_s_prime + q_l_prime
        discr = b_coeff ** 2 - 4 * alpha * (alpha + 1) * (1 + q_s_prime ** 2)
        z_sol = np.array(
            (
                (b_coeff[0] + np.sqrt(discr[0])) / (2 * (alpha + 1)),
                (b_coeff[1] - np.sqrt(discr[1])) / (2 * (alpha + 1)),
            )
        )
        qxl1 = z_sol / (1 - k ** 2)
        qxl2 = z_sol * (1 + q_l_prime ** 2) / (alpha * (1 + (q_s_prime - z_sol) ** 2))
        l_sol1 = qxl1 * np.real(self.z_src) / (2 * np.pi * self.f_c)
        l_sol2 = qxl2 * np.real(self.z_ld) / (2 * np.pi * self.f_c)
        return l_sol1, l_sol2

    def __enforce_symmetrical(self, _X_val, _of_load=True, sol=0):
        """
        return the 'distance' to a symmetrical balun (ie. primary = secondary)
        if _of_load, altering the load reactance
        else altering the source reactance
        """
        X_add = ((1 - sol) * _X_val, sol * _X_val)
        if _of_load:
            l1, l2 = self.design(XL_add=X_add)
        else:
            l1, l2 = self.design(XS_add=X_add)
        return float(np.abs(l1[sol] - l2[sol])[0])

    def enforce_symmetrical(self, side="load", _verbose=False):
        """
        return the reactance to be added to the load (if side="load") or the source impedance
        in order to realize a symmetrical balun (ie. primary = secondary)
        the two solution match the two solutions of the design
        """
        if side == "load":
            old_z = self.z_ld
            _through_load = True
        else:
            old_z = self.z_src
            _through_load = False
        # TODO : Detect when minimize did not converge
        res = list((0, 1))
        for i in (0, 1):
            t = minimize(
                self.__enforce_symmetrical,
                quality_f(old_z),
                args=(_through_load, i),
                method="Nelder-Mead",
            )
            res[i] = t.x[0]
        if _verbose:
            print(f"X_{side} must be change by {res[0]:5.2f} or {res[1]:5.2f}")
        return res



"""
Some function to format physical dimensions.
"""

from matplotlib.ticker import EngFormatter
from numpy import real, imag

"""
Format frequency to string.
"""
Frequency = EngFormatter(unit="Hz")

__res = EngFormatter(unit="")


def Impedance(value):
    """
    Format complex impedance to string with engineer format.
    """
    if imag(value) < 0:
        return __res(real(value)) + " - j" + __res(-imag(value)) + r" $\Omega$"
    return __res(real(value)) + " + j" + __res(imag(value)) + r" $\Omega$"


"""
Format number to string in engineer format without unit.
"""
SI = EngFormatter(unit="")



"""

"""
from numpy import sqrt
from .physical_dimension import PhysicalDimension
from .constants import c0


class Frequency(PhysicalDimension):
    unit = "Hz"

    def __init__(self, scale="lin", **data):
        super().__init__(scale=scale, **data)

    def to_time(self):
        if self.scale == "lin":
            return Time(value=1 / self.value, scale=self.scale)
        return Time(value=-self.value, scale=self.scale)

    def to_wavelength(self, eps_r: float = 1):
        f_lin = self.lin()
        return PhysicalDimension(
            value=sqrt(eps_r) * c0 / f_lin.value, scale="lin", unit="m"
        )


class Time(PhysicalDimension):
    unit = "s"

    def __init__(self, scale="lin", **data):
        super().__init__(scale=scale, **data)

    def to_freq(self):
        if self.scale == "lin":
            return Frequency(value=1 / self.value, scale=self.scale)
        return Frequency(value=-self.value, scale=self.scale)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['passive_auto_design',
 'passive_auto_design.components',
 'passive_auto_design.devices',
 'passive_auto_design.system',
 'passive_auto_design.units',
 'passive_auto_design.wrappers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'flake8>=4.0.1,<5.0.0',
 'matplotlib>=3.5.2,<4.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pytest-cov>=3.0.0,<4.0.0',
 'scikit-rf>=0.22.1,<0.24.0',
 'scipy>=1.8.0,<2.0.0']

setup_kwargs = {
    'name': 'passive-auto-design',
    'version': '0.1.10',
    'description': '',
    'long_description': None,
    'author': 'Patarimi',
    'author_email': '38954040+Patarimi@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""

"""
from numpy import pi, sqrt
from ..units.unit import SI


class Coupler:
    """
    Create a coupler object
    """

    def __init__(self, _fc=1e9, _zc=50, _k=0.707):
        self.f_c = _fc
        self.z_c = _zc
        self.k = _k
        self.l = self.z_c / (2 * pi * self.f_c * sqrt(1 - self.k ** 2))
        self.c = self.l / self.z_c ** 2

    def __str__(self):
        """
        Print a summary of the solution l and c
        """
        message = f"L: {SI(self.l)}H\tC: {SI(self.c)}F"
        return message



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from pydantic import BaseModel
from typing import Dict
from scipy.optimize import minimize_scalar
from matplotlib.ticker import EngFormatter
from  ..units.constants import u0, eps0


class LumpedElement(BaseModel, ABC):
    """
    class of standard lumped element, to be inherited by all lumped elements
    """

    dim: Dict[str, float]
    const: Dict[str, float]
    model: Dict[str, float] = {}

    def __init__(self, dim, const, **data):
        BaseModel.__init__(self, dim=dim, const=const, **data)
        self.model = self.get_model()

    # TODO enable multiple x_keys
    def set_model_with_dim(self, target_model: Dict[str, float], dim_key: str) -> float:
        """
        set the value of the dim_key to achieve the target model
        """
        res = minimize_scalar(self.__cost, args=(dim_key, target_model))
        self.model = self.get_model()
        return res.x

    def __cost(
        self, x_value: float, x_key: str, target_model: Dict[str, float]
    ) -> float:
        if x_key in self.dim:
            self.dim[x_key] = x_value
        if x_key in self.const:
            self.const[x_key] = x_value
        tmp_model = self.get_model()
        cost = 0
        for key in target_model.keys():
            cost += (target_model[key] - tmp_model[key]) ** 2 / (
                target_model[key] + tmp_model[key]
            ) ** 2
        return cost

    @abstractmethod
    def get_model(self):
        """
        Definition of the behavioral equation of the lumped element.
        Returns
        -------
        Dict[str, float]: list of the electrical parameter of the lumped element model
        """


Res = EngFormatter(unit=r"$\Omega$")


class Resistor(LumpedElement):
    """
    class describing a resistor behavior
    """

    def __init__(self, section=1e-3, length=1, rho=1e-15):
        param = {"section": section, "length": length}
        const = {"rho": rho}
        LumpedElement.__init__(self, dim=param, const=const)

    def __str__(self):
        return Res(self.model["res"])

    def get_model(self):
        return {
            "res": max(
                [self.const["rho"] * self.dim["length"] / self.dim["section"], 0.0]
            )
        }


Cap = EngFormatter(unit="F")


class Capacitor(LumpedElement):
    """
    class describing a capacitor behavior
    - dimension:
        - area: area of plate
        - dist: distance between plates
    - constant:
        - eps_r: relative permittivity of isolator
    - behavioral:
        - cap: capacitor
    """

    def __init__(self, area=1e-6, dist=1e-3, eps_r=1):
        dim = {"area": area, "dist": dist}
        const = {"eps_r": eps_r}
        LumpedElement.__init__(self, dim=dim, const=const)

    def __str__(self):
        return Cap(self.model["cap"])

    def get_model(self):
        return {"cap": eps0 * self.const["eps_r"] * self.dim["area"] / self.dim["dist"]}


Ind = EngFormatter(unit="H")


class Inductor(LumpedElement):
    """
    class describing a inductor behavior using wheeler formula for ic
    """

    def __init__(self, d_i=100e-6, n_turn=1, width=3e-6, gap=1e-6, k_1=2.25, k_2=3.55):
        param = {"d_i": d_i, "n_turn": n_turn, "width": width, "gap": gap}
        const = {"k_1": k_1, "k_2": k_2}
        LumpedElement.__init__(self, dim=param, const=const)

    def __str__(self):
        return Ind(self.model["ind"])

    def get_model(self):
        n = self.dim["n_turn"]
        outer_diam = (
            self.dim["d_i"] + 2 * n * self.dim["width"] + 2 * (n - 1) * self.dim["gap"]
        )
        self.dim["d_o"] = outer_diam
        rho = (self.dim["d_i"] + outer_diam) / 2
        density = (outer_diam - self.dim["d_i"]) / (outer_diam + self.dim["d_i"])
        ind = self.const["k_1"] * u0 * n ** 2 * rho / (1 + self.const["k_2"] * density)
        return {"ind": ind}
